/**************************************************************************
 * HASM64                                                                 *
 **************************************************************************/

/**************************************************************************
 * File name:     hasm64.c                                                *
 * Created:       1997-09-04 by Hampa Hug <hhug@student.ethz.ch>          *
 * Last modified: 2001-05-31 by Hampa Hug <hhug@student.ethz.ch>          *
 * Copyright:     (C) 1997-2001 by Hampa Hug <hhug@student.ethz.ch>       *
 **************************************************************************/

/**************************************************************************
 * This program is free software; you can redistribute it and/or modify   *
 * it under the terms of the GNU General Public License as published by   *
 * the Free Software Foundation; either version 2 of the License, or      *
 * (at your option) any later version.                                    *
 **************************************************************************/

#include <stdio.h>

#include "hasm64.h"


static int h64_scan = 0;
static int h64_verbose = 0;
static char *h64_symtab = NULL;


void prt_help (void)
{
  fprintf (stderr,
    "Usage: hasm64 options\n\n"
    "Options are:\n"
    "  -i string   Set input file name [stdin]\n"
    "  -o string   Set output file name [stdout]\n"
    "  -s, --scan  Scan only\n"
  );
  fflush (stderr);
}

void prt_version (void)
{
  fprintf (stderr, "hasm64 version 0.0.2\n");
  fflush (stdout);
}

void prt_error (char *msg)
{
  fprintf (stderr, "hasm64: %s\n", msg);
  fflush (stderr);
}

char *str_clone (char *str)
{
  char *ret;

  ret = (char *) malloc (strlen (str) + 1);
  if (ret == NULL) {
    return (NULL);
  }

  strcpy (ret, str);

  return (ret);
}

char *str_copy (char *dst, char *src, unsigned n)
{
  unsigned i;

  if (n > 0) {
    n -= 1;
  }

  i = 0;
  while ((i < n) && (src[i] != 0)) {
    dst[i] = src[i];
    i += 1;
  }

  dst[i] = 0;

  return (dst);
}

int str_isarg1 (char *str, char *arg)
{
  if (strcmp (str, arg) == 0) {
    return (1);
  }

  return (0);
}

int str_isarg2 (char *str, char *arg1, char *arg2)
{
  if (strcmp (str, arg1) == 0) {
    return (1);
  }

  if (strcmp (str, arg2) == 0) {
    return (1);
  }

  return (0);
}

void err_print (parser_t *par, unsigned err, char *str)
{
  unsigned long ln;
  char          *tok, *fname;

  ln = scn_get_line (par->scn) + 1;
  tok = par->scn->tok_str;

  fname = scn64_get_file (par->scn);
  if (fname == NULL) {
    fname = "-";
  }

  if (str == NULL) {
    str = "Unknown error";
  }

  fprintf (stderr, "%s:%lu: (%s): errror %d: %s\n", fname, ln, tok, err, str);
  fflush (stderr);
}

int hasm64_scan (char *f_inp, char *f_out)
{
  scanner_t *scn;
  FILE      *fp;

  scn = scn64_new ();
  if (scn64_set_file (scn, f_inp)) {
    return (1);
  }

  fp = fopen (f_out, "wb");
  if (fp == NULL) {
    scn->del (scn);
    return (1);
  }

  scn->scan (scn);

  while (scn->tok_id != SCN_TOK_NONE) {
    if (scn->tok_id == TOK_STRING) {
      fprintf (fp, "\"%s\"\n", scn->tok_str);
    }
    else {
      fprintf (fp, "%s\n", scn->tok_str);
    }

    scn->scan (scn);
  }

  fclose (fp);

  scn->del (scn);

  return (0);
}

int hasm64_asm (char *f_inp, char *f_out)
{
  scanner_t *scn;
  parser_t  *par;

  par = par_new ();

  do {
    prtv_msg_1 ("Pass %u\n", par->pass);

    scn = scn64_new ();
    if (scn == NULL) {
      par_del (par);
      return (1);
    }

    if (scn64_set_file (scn, f_inp)) {
      scn->del (scn);
      par_del (par);
      return (1);
    }

    par->newpass = 0;

    par_set_scn (par, scn);
    par_scan (par);

    cg_set_addr (par->cg, 0);
    cg_clear (par->cg);

    if (par_module (par)) {
      par_del (par);

      return (1);
    }

    par->pass += 1;

    par_set_scn (par, NULL);
  } while (par->newpass);

  cg_write_file (par->cg, f_out);

  if (h64_symtab != NULL) {
    st_dump_file (par->st, h64_symtab);
  }

  par_del (par);

  return (0);
}

int main (int argc, char **argv)
{
  int       i, r;
  char      *f_inp, *f_out;

  if (argc == 2) {
    if (str_isarg1 (argv[1], "--help")) {
      prt_help ();
      return (0);
    }
    else if (str_isarg1 (argv[1], "--version")) {
      prt_version ();
      return (0);
    }
  }

  f_inp = NULL;
  f_out = NULL;

  i = 1;
  while (i < argc) {
    if (str_isarg2 (argv[i], "-v", "--verbose")) {
      h64_verbose = 1;
    }
    else if (str_isarg1 (argv[i], "-i")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing input file name");
        return (1);
      }

      f_inp = str_clone (argv[i]);
    }
    else if (str_isarg1 (argv[i], "-o")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing output file name");
        return (1);
      }

      f_out = str_clone (argv[i]);
    }
    else if (str_isarg2 (argv[i], "-d", "--symtab")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing symtab file name");
        return (1);
      }

      h64_symtab = str_clone (argv[i]);
    }
    else if (str_isarg2 (argv[i], "-s", "--scan")) {
      h64_scan = 1;
    }
    else {
      prt_error ("Unknown option");
      return (1);
    }

    i += 1;
  }

  if (h64_scan) {
    r = hasm64_scan (f_inp, f_out);
  }
  else {
    r = hasm64_asm (f_inp, f_out);
  }

  return (r);
}
